﻿using System;
using System.Collections.Generic;
using System.Text;

using ThinkGeo.MapSuite.Core;


namespace TestKmlMapSuiteCore
{
    class KmlGeoCanvas : GeoCanvas
    {
        StringBuilder kmlBuilder;

        private const float virtualMapWidth = 100f;
        private const float virtualMapHeight = 100f;
        private RectangleShape virtualWorldExtent = new RectangleShape();

        private Dictionary<int, GeoBrush> geoBrushDictionary = new Dictionary<int, GeoBrush>();

        private Dictionary<int, string> styleUrlDictionary = new Dictionary<int, string>();

        private string extrudeString = string.Empty;
        private string tessellateString = string.Empty;
        private string altitudeModeString = string.Empty;

        private StringBuilder contentStringBuildLevel1 = new StringBuilder();
        private StringBuilder contentStringBuildLevel2 = new StringBuilder();
        private StringBuilder contentStringBuildLevel3 = new StringBuilder();
        private StringBuilder contentStringBuildLevel4 = new StringBuilder();


        public bool Extrude
        {
            get;
            set;
        }

        public bool Tessellate
        {
            get;
            set;
        }

        public AltitudeMode AltitudeMode
        {
            get;
            set;
        }

        protected override void BeginDrawingCore(object nativeImage, RectangleShape worldExtent, GeographyUnit drawingMapUnit)
        {
            kmlBuilder = (StringBuilder)nativeImage;

            virtualWorldExtent = worldExtent;

            extrudeString = Extrude ? @"<extrude>1</extrude>" : "";
            tessellateString = Tessellate ? @"<tessellate>1</tessellate>" : "";
            switch (AltitudeMode)
            {
                case AltitudeMode.Absolute:
                    altitudeModeString = "<altitudeMode>absolute</altitudeMode>";
                    break;
                case AltitudeMode.RelativeToGround:
                    altitudeModeString = "<altitudeMode>relativeToGround</altitudeMode>";
                    break;
                default:
                    break;
            }

            kmlBuilder.AppendLine(@"<?xml version=""1.0"" encoding=""UTF-8""?>");
            kmlBuilder.AppendLine(@"<kml xmlns=""http://www.opengis.net/kml/2.2"">");
            kmlBuilder.AppendLine(@"<Document>");
        }

        protected override void DrawAreaCore(IEnumerable<ScreenPointF[]> screenPoints, GeoPen outlinePen, GeoBrush fillBrush, DrawingLevel drawingLevel, float xOffset, float yOffset, PenBrushDrawingOrder penBrushDrawingOrder)
        {
            if (fillBrush == null)
            {
                fillBrush = new GeoSolidBrush(GeoColor.SimpleColors.Transparent);
            }

            int id = 0;
            if (outlinePen != null)
            {
                id = outlinePen.GetHashCode();
            }
            else if (fillBrush != null)
            {
                id = id ^ fillBrush.GetHashCode();
            }

            if (!styleUrlDictionary.ContainsKey(id))
            {
                string kmlStyle = GetPolygonStyleKml(id, outlinePen, ((GeoSolidBrush)fillBrush).Color);
                kmlBuilder.Append(kmlStyle);
                styleUrlDictionary.Add(id, string.Format("<styleUrl>#{0}</styleUrl>", id));
            }

            StringBuilder contentStringBuilder = GetStringBuilder(drawingLevel);

            contentStringBuilder.AppendLine();
            contentStringBuilder.AppendLine(@"<Placemark>");
            contentStringBuilder.AppendLine(styleUrlDictionary[id]);
            contentStringBuilder.AppendLine(@"<Polygon>");

            contentStringBuilder.AppendLine(extrudeString);
            contentStringBuilder.AppendLine(tessellateString);
            contentStringBuilder.AppendLine(altitudeModeString);

            bool firstCoordinates = true;
            foreach (ScreenPointF[] screenPoint in screenPoints)
            {
                if (firstCoordinates)
                {
                    contentStringBuilder.AppendLine(@"<outerBoundaryIs>");
                    AppendLinearRing(screenPoint, xOffset, yOffset, contentStringBuilder);
                    contentStringBuilder.AppendLine(@"</outerBoundaryIs>");
                    firstCoordinates = false;
                }
                else
                {
                    contentStringBuilder.AppendLine(@"<innerBoundaryIs>");
                    AppendLinearRing(screenPoint, xOffset, yOffset, contentStringBuilder);
                    contentStringBuilder.AppendLine(@"</innerBoundaryIs>");
                }
            }
            contentStringBuilder.AppendLine(@"</Polygon>");
            contentStringBuilder.AppendLine(@"</Placemark>");
        }

        protected override void DrawEllipseCore(ScreenPointF screenPoint, float width, float height, GeoPen outlinePen, GeoBrush fillBrush, DrawingLevel drawingLevel, float xOffset, float yOffset, PenBrushDrawingOrder penBrushDrawingOrder)
        {
            throw new NotImplementedException();
        }

        protected override void DrawLineCore(IEnumerable<ScreenPointF> screenPoints, GeoPen linePen, DrawingLevel drawingLevel, float xOffset, float yOffset)
        {
            int id = linePen.GetHashCode();
            if (!styleUrlDictionary.ContainsKey(id))
            {
                string kmlStyle = GetLineStyleKml(id, linePen);
                kmlBuilder.Append(kmlStyle);
                styleUrlDictionary.Add(id, string.Format("<styleUrl>#{0}</styleUrl>", id));
            }

            StringBuilder contentStringBuilder = GetStringBuilder(drawingLevel);

            contentStringBuilder.AppendLine();
            contentStringBuilder.AppendLine(@"<Placemark>");
            contentStringBuilder.AppendLine(styleUrlDictionary[id]);
            contentStringBuilder.AppendLine(@"<LineString>");

            contentStringBuilder.AppendLine(extrudeString);
            contentStringBuilder.AppendLine(tessellateString);
            contentStringBuilder.AppendLine(altitudeModeString);

            AppendCoordinates(screenPoints, xOffset, yOffset, contentStringBuilder);
            contentStringBuilder.AppendLine(@"</LineString>");
            contentStringBuilder.AppendLine(@"</Placemark>");
        }

        private void AppendCoordinates(IEnumerable<ScreenPointF> screenPoints, float xOffset, float yOffset, StringBuilder contentStringBuilder)
        {
            contentStringBuilder.AppendLine(@"<coordinates>");

            foreach (ScreenPointF screenPoint in screenPoints)
            {
                PointShape pointShape = ExtentHelper.ToWorldCoordinate(virtualWorldExtent, screenPoint.X + xOffset, screenPoint.Y + yOffset, virtualMapWidth, virtualMapHeight);
                contentStringBuilder.AppendFormat(" {0},{1},{2} ", pointShape.X, pointShape.Y, 500);
            }
            contentStringBuilder.AppendLine(@"</coordinates>");
        }

        private void AppendLinearRing(IEnumerable<ScreenPointF> screenPoints, float xOffset, float yOffset, StringBuilder contentStringBuilder)
        {
            contentStringBuilder.AppendLine(@"<LinearRing>");
            AppendCoordinates(screenPoints, xOffset, yOffset, contentStringBuilder);
            contentStringBuilder.AppendLine(@"</LinearRing>");
        }

        private StringBuilder GetStringBuilder(DrawingLevel drawingLevel)
        {
            StringBuilder result = new StringBuilder();
            switch (drawingLevel)
            {
                case DrawingLevel.LevelOne:
                    result = contentStringBuildLevel1;
                    break;
                case DrawingLevel.LevelTwo:
                    result = contentStringBuildLevel2;
                    break;
                case DrawingLevel.LevelThree:
                    result = contentStringBuildLevel3;
                    break;
                case DrawingLevel.LevelFour:
                    result = contentStringBuildLevel4;
                    break;
            }
            return result;
        }

        private string GetLineStyleKml(int id, GeoPen linePen)
        {
            StringBuilder styleBuilder = new StringBuilder();
            styleBuilder.AppendLine();
            styleBuilder.AppendFormat(@"<Style id=""{0}"">", id);
            styleBuilder.AppendLine();
            styleBuilder.AppendLine(@"<LineStyle>");
            styleBuilder.AppendFormat(@"<color>{0}</color>", GetGoogleHTMLColor(linePen.Color));
            styleBuilder.AppendLine();
            styleBuilder.AppendFormat(@"<width>{0}</width>", linePen.Width);
            styleBuilder.AppendLine();
            styleBuilder.AppendLine(@"</LineStyle>");
            styleBuilder.AppendLine(@"</Style>");

            return styleBuilder.ToString();
        }

        private string GetPolygonStyleKml(int id, GeoPen outlinePen, GeoColor fillColor)
        {
            StringBuilder styleBuilder = new StringBuilder();
            styleBuilder.AppendLine();
            styleBuilder.AppendFormat(@"<Style id=""{0}"">", id);
            styleBuilder.AppendLine();
            styleBuilder.AppendLine(@"<LineStyle>");
            if (outlinePen != null)
            {
                styleBuilder.AppendFormat(@"<color>{0}</color>", GetGoogleHTMLColor(outlinePen.Color));
                styleBuilder.AppendLine();
                styleBuilder.AppendFormat(@"<width>{0}</width>", outlinePen.Width);
            }
            styleBuilder.AppendLine();
            styleBuilder.AppendLine(@"</LineStyle>");
            styleBuilder.AppendLine(@"<PolyStyle>");
            styleBuilder.AppendFormat(@"<color>{0}</color>", GetGoogleHTMLColor(fillColor));
            styleBuilder.AppendLine();
            styleBuilder.AppendLine(@"</PolyStyle>");

            styleBuilder.AppendLine(@"</Style>");

            return styleBuilder.ToString();
        }


        private string GetGoogleHTMLColor(GeoColor geoColor)
        {
            StringBuilder googleHtmlColor = new StringBuilder();
            googleHtmlColor.Append(GetColorComponentInHex(geoColor.AlphaComponent));
            googleHtmlColor.Append(GetColorComponentInHex(geoColor.BlueComponent));
            googleHtmlColor.Append(GetColorComponentInHex(geoColor.GreenComponent));
            googleHtmlColor.Append(GetColorComponentInHex(geoColor.RedComponent));
            return googleHtmlColor.ToString();
        }

        private string GetColorComponentInHex(byte component)
        {
            return Convert.ToInt32(component).ToString("X").PadLeft(2, '0');
        }

        protected override void DrawScreenImageCore(GeoImage image, float centerXInScreen, float centerYInScreen, float widthInScreen, float heightInScreen, DrawingLevel drawingLevel, float xOffset, float yOffset, float rotateAngle)
        {
            throw new NotImplementedException();
        }

        protected override void DrawScreenImageWithoutScalingCore(GeoImage image, float centerXInScreen, float centerYInScreen, DrawingLevel drawingLevel, float xOffset, float yOffset, float rotateAngle)
        {
            throw new NotImplementedException();
        }

        protected override void DrawTextCore(string text, GeoFont font, GeoBrush fillBrush, GeoPen haloPen, IEnumerable<ScreenPointF> textPathInScreen, DrawingLevel drawingLevel, float xOffset, float yOffset, float rotateAngle)
        {
//            throw new NotImplementedException();
        }

        protected override float GetCanvasHeightCore(object nativeImage)
        {
            return virtualMapHeight;
        }

        protected override float GetCanvasWidthCore(object nativeImage)
        {
            return virtualMapWidth;
        }

        public override System.IO.Stream GetStreamFromGeoImage(GeoImage image)
        {
            throw new NotImplementedException();
        }

        protected override DrawingRectangleF MeasureTextCore(string text, GeoFont font)
        {
            throw new NotImplementedException();
        }

        protected override GeoImage ToGeoImageCore(object nativeImage)
        {
            throw new NotImplementedException();
        }

        protected override object ToNativeImageCore(GeoImage image)
        {
            throw new NotImplementedException();
        }

        protected override void EndDrawingCore()
        {
            kmlBuilder.Append(contentStringBuildLevel4.ToString());
            kmlBuilder.Append(contentStringBuildLevel3.ToString());
            kmlBuilder.Append(contentStringBuildLevel2.ToString());
            kmlBuilder.Append(contentStringBuildLevel1.ToString());
            kmlBuilder.AppendFormat(@"</Document>");
            kmlBuilder.AppendFormat(@"</kml>");
        }

        protected override void FlushCore()
        {
            throw new NotImplementedException();
        }
    }
}
